import fnmatch
import re
from dataclasses import dataclass
from typing import Union, Optional, List, Tuple
from urllib.parse import quote_plus

from .http import get_session, ENDPOINT


class ModelNotFound(Exception):
    pass


class ModelFoundDuplicated(Exception):
    pass


class ModelVersionNotFound(Exception):
    pass


class ModelVersionDuplicated(Exception):
    pass


class ResourceNotFound(Exception):
    pass


class ResourceDuplicated(Exception):
    pass


class Model:
    def __init__(self, data):
        self.data = data

    @property
    def model_name(self) -> str:
        return self.data['name']

    @property
    def model_id(self) -> int:
        return self.data['id']

    @property
    def creator(self) -> str:
        return self.data['creator']['username']

    def __repr__(self):
        return (f'<{self.__class__.__name__} name: {self.model_name!r}, '
                f'id: {self.model_id!r}, creator: {self.creator!r}>')


@dataclass
class Resource:
    """
    Data class for resource.
    """
    model_name: str
    model_id: int
    creator: str
    version_name: str
    version_id: int
    filename: str
    is_primary: bool
    url: str
    sha256: str
    crc32: str
    hashes: dict
    size: int
    tags: List[str]


def find_model_by_id(model_id) -> dict:
    """
    Retrieve model information from the CiviTAI API based on the given model ID.

    :param model_id: The ID of the model to retrieve information for.
    :type model_id: int
    :return: The dictionary containing the model information.
    :rtype: dict
    :raises ModelNotFound: If the model with the given ID is not found.
    """
    resp = get_session().get(f'{ENDPOINT}/api/v1/models/{model_id}')
    if resp.ok:
        return resp.json()
    else:
        if resp.status_code == 404:
            raise ModelNotFound(model_id)
        else:
            resp.raise_for_status()


def _name_strip(name: str) -> str:
    return re.sub(r'[\W_]+', '', name.lower())


def list_models_by_name(model_name: str, creator: Optional[str] = None, strict: bool = False) -> List[dict]:
    """
    Retrieve model information from the CiviTAI API based on the given model name.

    :param model_name: The name of the model to retrieve information for.
    :type model_name: str
    :param creator: Name of creator. ``None`` means anyone.
    :type creator: Optional[str]
    :param strict: Strict filter all the results or not. Default is ``False``.
    :type strict: bool
    :return: The list of dictionaries containing the searched model information.
    :rtype: dict
    """
    _session = get_session()
    params = {'query': model_name}
    if creator:
        params['username'] = creator

    resp = _session.get(f'{ENDPOINT}/api/v1/models', params=params)
    resp.raise_for_status()
    collected_items = []
    for item in resp.json()['items']:
        if not strict or (_name_strip(item['name']) == _name_strip(model_name) and
                          (creator is None or _name_strip(item['creator']['username']) == _name_strip(creator))):
            collected_items.append(item)

    return collected_items


def find_model_by_name(model_name: str, creator: Optional[str] = None) -> dict:
    """
    Retrieve model information from the CiviTAI API based on the given model name.

    :param model_name: The name of the model to retrieve information for.
    :type model_name: str
    :param creator: Name of creator. ``None`` means anyone.
    :type creator: Optional[str]
    :return: The dictionary containing the model information.
    :rtype: dict
    :raises ModelNotFound: If the model with the given name is not found.
    :raises ModelFoundDuplicated: If multiple models with the same name are found.
    """
    collected_items = list_models_by_name(model_name, creator, strict=True)
    if not collected_items:
        raise ModelNotFound(model_name)
    elif len(collected_items) > 1:
        all_names = [item['name'] for item in collected_items]
        raise ModelFoundDuplicated(model_name, all_names)
    else:
        return collected_items[0]


def _maybe_a_hash(text: str):
    return isinstance(text, str) and re.fullmatch(r'^[\dA-F]{8,}$', text.upper())


def find_version_id_by_hash(model_hash: str) -> Optional[Tuple[int, int, str]]:
    if not _maybe_a_hash(model_hash):
        return None

    resp = get_session().get(f'{ENDPOINT}/api/v1/model-versions/by-hash/{quote_plus(model_hash.upper())}')
    if resp.status_code == 404:
        return None
    else:
        resp.raise_for_status()
        data = resp.json()
        version_file = None
        for file in data['files']:
            if model_hash.upper() in set((file.get('hashes') or {}).values()):
                version_file = file['name']
                break

        assert version_file is not None, f'No file in model version {data["id"]!r} ' \
                                         f'matches the given hash {model_hash!r}.'
        return data['modelId'], data['id'], version_file


def find_model(model_name_or_id: Union[int, str], creator: Optional[str] = None) -> dict:
    """
    Find model information from the CiviTAI API based on the given model name or ID.

    :param model_name_or_id: The name or ID of the model to retrieve information for.
    :type model_name_or_id: Union[int, str]
    :param creator: Name of creator. ``None`` means anyone.
    :type creator: Optional[str]
    :return: The dictionary containing the model information.
    :rtype: dict
    :raises TypeError: If the model name or ID is not a valid integer or string.
    """
    if isinstance(model_name_or_id, int):
        return find_model_by_id(model_name_or_id)
    elif isinstance(model_name_or_id, str):
        if creator is None:
            try:
                model_id = int(model_name_or_id)
                return find_model_by_id(model_id)
            except (ModelNotFound, TypeError, ValueError):
                pass
        return find_model_by_name(model_name_or_id, creator)
    else:
        raise TypeError(f'Unknown model name or id, it should be an integer or string - {model_name_or_id!r}.')


def find_version(model_data: dict, version: Union[int, str, None] = None):
    """
    Find the specified version from the model data.

    :param model_data: The model data containing version information.
    :type model_data: dict
    :param version: The version ID or name to find. If None, the first version will be chosen.
    :type version: Union[int, str, None]
    :return: The dictionary containing the selected version information.
    :rtype: dict
    :raises ModelVersionNotFound: If the specified version is not found.
    :raises ModelVersionDuplicated: If multiple versions with the same ID or name are found.
    """
    # find chosen version
    versions = model_data['modelVersions']
    if version is None:
        select_version = versions[0]
    else:
        all_select_versions = []
        for vitem in versions:
            if vitem['id'] == version or _name_strip(vitem['name']) == _name_strip(str(version)):
                all_select_versions.append(vitem)

        if not all_select_versions:
            raise ModelVersionNotFound(model_data['name'], version)
        elif len(all_select_versions) > 1:
            raise ModelVersionDuplicated(model_data['name'], [vitem['name'] for vitem in all_select_versions])
        else:
            select_version = all_select_versions[0]

    return select_version


def find_resource(model_data: dict, version_data: dict, pattern: str = None):
    """
    Find the specified resource from the model and version data.

    :param model_data: The model data containing version information.
    :type model_data: dict
    :param version_data: The version data containing resource information.
    :type version_data: dict
    :param pattern: The pattern to match the resource filename. If None, the primary resource will be chosen.
    :type pattern: str
    :return: The resource information.
    :rtype: Resource
    :raises TypeError: If the pattern of the resource is not a valid string.
    :raises ResourceNotFound: If the specified resource is not found.
    :raises ResourceDuplicated: If multiple resources with the same name are found.
    """
    # find chosen resource
    if pattern is None:
        pattern, primary = '*', True
    elif isinstance(pattern, str):
        primary = False
    else:
        raise TypeError(f'Pattern of resource should be a string, but {pattern!r} found.')

    all_select_files = []
    for file in version_data['files']:
        if (primary and file.get('primary')) or (not primary and fnmatch.fnmatch(file['name'], pattern)):
            all_select_files.append(file)
    if not all_select_files:
        raise ResourceNotFound(model_data['name'], version_data['name'], {'pattern': pattern, 'primary': primary})
    elif len(all_select_files) > 1:
        raise ResourceDuplicated(model_data['name'], version_data['name'], [file['name'] for file in all_select_files])
    else:
        select_file = all_select_files[0]

    file_size = select_file['sizeKB'] * 1024
    assert abs(round(file_size) - file_size) < 1e-4
    file_size = int(round(file_size))
    return Resource(
        model_name=model_data['name'], model_id=model_data['id'],
        creator=model_data['creator']['username'],
        version_name=version_data['name'], version_id=version_data['id'],
        filename=select_file['name'],
        url=select_file['downloadUrl'],
        sha256=select_file['hashes'].get('SHA256'),
        crc32=select_file['hashes'].get('CRC32'),
        hashes=select_file['hashes'],
        is_primary=select_file.get('primary', False),
        size=file_size,
        tags=model_data['tags'],
    )



import os
from unittest import skipUnless
from unittest.mock import patch

import pytest
from hbutils.testing import simulate_entry, isolated_directory

from pycivitai.config.meta import __TITLE__, __VERSION__
from pycivitai.entry import cli
from pycivitai.manager import DispatchManager
from .testings import get_testfile

try:
    import InquirerPy
except (ModuleNotFoundError, ImportError):
    InquirerPy = None


@pytest.fixture()
def sample_repo_dir():
    return get_testfile('sample_repo_1')


@pytest.fixture()
def sample_repo(sample_repo_dir):
    with isolated_directory({'repo': sample_repo_dir}):
        def _get_manager(offline):
            return DispatchManager('repo', offline)

        with patch('pycivitai.entry._get_global_manager', _get_manager), \
                patch('pycivitai.dispatch._get_global_manager', _get_manager):
            yield DispatchManager('repo', offline=True)


@pytest.fixture()
def empty_repo():
    with isolated_directory():
        os.makedirs('repo', exist_ok=True)

        def _get_manager(offline):
            return DispatchManager('repo', offline)

        with patch('pycivitai.entry._get_global_manager', _get_manager), \
                patch('pycivitai.dispatch._get_global_manager', _get_manager):
            yield DispatchManager('repo', offline=True)


@pytest.fixture()
def mock_all_choices():
    from InquirerPy.base import Choice
    def _get_choices(choices):
        retval = []
        for item in choices:
            if isinstance(item, Choice):
                retval.append(item.value)

        return retval

    with patch('pycivitai.entry._get_choices', _get_choices):
        yield


@pytest.fixture()
def mock_no_choices():
    def _get_choices(choices):
        return []

    with patch('pycivitai.entry._get_choices', _get_choices):
        yield


@pytest.mark.unittest
class TestEntry:
    def test_version(self):
        result = simulate_entry(cli, ['cli', '-v'])
        assert result.exitcode == 0, f'Exitcode - {result.exitcode}\n' \
                                     f'Stdout:\n' \
                                     f'{result.stdout}\n' \
                                     f'\n' \
                                     f'Stderr:\n' \
                                     f'{result.stderr}'
        assert __TITLE__.lower() in result.stdout.lower()
        assert __VERSION__.lower() in result.stdout.lower()

    def test_delete_all(self, sample_repo, text_aligner):
        result = simulate_entry(cli, ['cli', 'delete-cache', '-A'])
        assert result.exitcode == 0, f'Exitcode - {result.exitcode}\n' \
                                     f'Stdout:\n' \
                                     f'{result.stdout}\n' \
                                     f'\n' \
                                     f'Stderr:\n' \
                                     f'{result.stderr}'
        text_aligner.assert_equal(
            'All models deleted, total 24.854 KiB.',
            result.stdout,
        )

    @skipUnless(InquirerPy is not None, 'InquirePy required')
    def test_delete_with_tui(self, sample_repo, text_aligner, mock_all_choices):
        assert sample_repo.total_size == 25451
        with patch('pycivitai.entry._confirm', lambda x: True):
            result = simulate_entry(cli, ['cli', 'delete-cache'])
            assert result.exitcode == 0, f'Exitcode - {result.exitcode}\n' \
                                         f'Stdout:\n' \
                                         f'{result.stdout}\n' \
                                         f'\n' \
                                         f'Stderr:\n' \
                                         f'{result.stderr}'
            text_aligner.assert_equal(
                'Deletion complete!',
                result.stdout,
            )
        assert sample_repo.total_size == 0

    @skipUnless(InquirerPy is not None, 'InquirePy required')
    def test_delete_with_tui_cancelled(self, sample_repo, text_aligner, mock_all_choices):
        assert sample_repo.total_size == 25451
        with patch('pycivitai.entry._confirm', lambda x: False):
            result = simulate_entry(cli, ['cli', 'delete-cache'])
            assert result.exitcode == 0, f'Exitcode - {result.exitcode}\n' \
                                         f'Stdout:\n' \
                                         f'{result.stdout}\n' \
                                         f'\n' \
                                         f'Stderr:\n' \
                                         f'{result.stderr}'
        assert sample_repo.total_size == 25451

    @skipUnless(InquirerPy is not None, 'InquirePy required')
    def test_delete_with_tui_nothing(self, empty_repo, text_aligner, mock_all_choices):
        assert empty_repo.total_size == 0
        with patch('pycivitai.entry._confirm', lambda x: True):
            result = simulate_entry(cli, ['cli', 'delete-cache'])
            assert result.exitcode == 0, f'Exitcode - {result.exitcode}\n' \
                                         f'Stdout:\n' \
                                         f'{result.stdout}\n' \
                                         f'\n' \
                                         f'Stderr:\n' \
                                         f'{result.stderr}'
            text_aligner.assert_equal(
                'No models found to delete.',
                result.stdout,
            )
        assert empty_repo.total_size == 0

    @skipUnless(InquirerPy is not None, 'InquirePy required')
    def test_delete_with_tui_no_choice(self, sample_repo, text_aligner, mock_no_choices):
        assert sample_repo.total_size == 25451
        with patch('pycivitai.entry._confirm', lambda x: True):
            result = simulate_entry(cli, ['cli', 'delete-cache'])
            assert result.exitcode == 0, f'Exitcode - {result.exitcode}\n' \
                                         f'Stdout:\n' \
                                         f'{result.stdout}\n' \
                                         f'\n' \
                                         f'Stderr:\n' \
                                         f'{result.stderr}'
            text_aligner.assert_equal(
                'Deletion cancelled.',
                result.stdout,
            )
        assert sample_repo.total_size == 25451

    def test_get(self, sample_repo):
        result = simulate_entry(cli, ['cli', 'get', '-m', 'amiya arknights (old)'])
        assert result.exitcode == 0, f'Exitcode - {result.exitcode}\n' \
                                     f'Stdout:\n' \
                                     f'{result.stdout}\n' \
                                     f'\n' \
                                     f'Stderr:\n' \
                                     f'{result.stderr}'
        assert os.path.samefile(
            result.stdout.strip(),
            os.path.join('repo', 'amiya_arknights_old__narugo1992__115427', 'v1_1__124885', 'files', 'amiya.pt'),
        )

    def test_get_offline(self, sample_repo):
        result = simulate_entry(cli, ['cli', 'get', '-m', 'amiya arknights (old)', '--offline'])
        assert result.exitcode == 0, f'Exitcode - {result.exitcode}\n' \
                                     f'Stdout:\n' \
                                     f'{result.stdout}\n' \
                                     f'\n' \
                                     f'Stderr:\n' \
                                     f'{result.stderr}'
        assert os.path.samefile(
            result.stdout.strip(),
            os.path.join('repo', 'amiya_arknights_old__narugo1992__115427', 'v1_0__124870', 'files', 'amiya.pt'),
        )

    def test_get_specific_version(self, sample_repo):
        result = simulate_entry(cli, ['cli', 'get', '-m', 'amiya arknights (old)', '-v', 'v1.0'])
        assert result.exitcode == 0, f'Exitcode - {result.exitcode}\n' \
                                     f'Stdout:\n' \
                                     f'{result.stdout}\n' \
                                     f'\n' \
                                     f'Stderr:\n' \
                                     f'{result.stderr}'
        assert os.path.samefile(
            result.stdout.strip(),
            os.path.join('repo', 'amiya_arknights_old__narugo1992__115427', 'v1_0__124870', 'files', 'amiya.pt'),
        )



from .http import get_session, configure_http_backend, ENDPOINT, OFFLINE_MODE, OfflineModeEnabled
from .resource import find_version, find_resource, Resource, ResourceNotFound, ModelNotFound, ModelVersionNotFound, \
    ResourceDuplicated, ModelVersionDuplicated, ModelFoundDuplicated, find_model, find_model_by_name, find_model_by_id, \
    find_version_id_by_hash, Model, list_models_by_name



import os
import threading
from functools import lru_cache
from typing import Callable

import requests


ENDPOINT = 'https://civitai.com'
OFFLINE_MODE = bool(os.environ.get('CIVITAI_OFFLINE'))

BACKEND_FACTORY_T = Callable[[], requests.Session]
_GLOBAL_BACKEND_FACTORY: BACKEND_FACTORY_T = requests.session


class OfflineModeEnabled(Exception):
    pass


def configure_http_backend(backend_factory: BACKEND_FACTORY_T = requests.Session) -> None:
    """
    Configure the HTTP backend by providing a `backend_factory`. Any HTTP calls made by `huggingface_hub` will use a
    Session object instantiated by this factory. This can be useful if you are running your scripts in a specific
    environment requiring custom configuration (e.g. custom proxy or certifications).

    Use [`get_session`] to get a configured Session. Since `requests.Session` is not guaranteed to be thread-safe,
    `huggingface_hub` creates 1 Session instance per thread. They are all instantiated using the same `backend_factory`
    set in [`configure_http_backend`]. A LRU cache is used to cache the created sessions (and connections) between
    calls. Max size is 128 to avoid memory leaks if thousands of threads are spawned.

    See [this issue](https://github.com/psf/requests/issues/2766) to know more about thread-safety in `requests`.

    Example:
    ```py
    import requests
    from huggingface_hub import configure_http_backend, get_session

    # Create a factory function that returns a Session with configured proxies
    def backend_factory() -> requests.Session:
        session = requests.Session()
        session.proxies = {"http": "http://10.10.1.10:3128", "https": "https://10.10.1.11:1080"}
        return session

    # Set it as the default session factory
    configure_http_backend(backend_factory=backend_factory)

    # In practice, this is mostly done internally in `huggingface_hub`
    session = get_session()
    ```
    """
    global _GLOBAL_BACKEND_FACTORY
    _GLOBAL_BACKEND_FACTORY = backend_factory
    _get_session_from_cache.cache_clear()


def get_session() -> requests.Session:
    """
    Get a `requests.Session` object, using the session factory from the user.

    Use [`get_session`] to get a configured Session. Since `requests.Session` is not guaranteed to be thread-safe,
    `huggingface_hub` creates 1 Session instance per thread. They are all instantiated using the same `backend_factory`
    set in [`configure_http_backend`]. A LRU cache is used to cache the created sessions (and connections) between
    calls. Max size is 128 to avoid memory leaks if thousands of threads are spawned.

    See [this issue](https://github.com/psf/requests/issues/2766) to know more about thread-safety in `requests`.

    Example:
    ```py
    import requests
    from huggingface_hub import configure_http_backend, get_session

    # Create a factory function that returns a Session with configured proxies
    def backend_factory() -> requests.Session:
        session = requests.Session()
        session.proxies = {"http": "http://10.10.1.10:3128", "https": "https://10.10.1.11:1080"}
        return session

    # Set it as the default session factory
    configure_http_backend(backend_factory=backend_factory)

    # In practice, this is mostly done internally in `huggingface_hub`
    session = get_session()
    ```
    """
    return _get_session_from_cache(thread_ident=threading.get_ident())


@lru_cache(maxsize=128)  # default value for Python>=3.8. Let's keep the same for Python3.7
def _get_session_from_cache(thread_ident: int) -> requests.Session:
    """
    Create a new session per thread using global factory. Using LRU cache (maxsize 128) to avoid memory leaks when
    using thousands of threads. Cache is cleared when `configure_http_backend` is called.
    """
    _ = thread_ident
    return _GLOBAL_BACKEND_FACTORY()



from .dispatch import DispatchManager, LocalModelNotFound, LocalModelDuplicated
from .model import ModelManager, LocalVersionNotFound, LocalVersionDuplicated
from .version import VersionManager, LocalFileDuplicated, LocalFileNotFound, LocalPrimaryFileUnset, LocalFile



import hashlib
import os

import pytest

from pycivitai.client import ResourceDuplicated, ModelFoundDuplicated
from pycivitai.dispatch import civitai_find_online, civitai_download, civitai_search_online


def calculate_sha256(file_path):
    sha256_hash = hashlib.sha256()
    block_size = 65536

    with open(file_path, 'rb') as file:
        while True:
            data = file.read(block_size)
            if not data:
                break
            sha256_hash.update(data)

    return sha256_hash.hexdigest().upper()


@pytest.mark.unittest
class TestDispatch:
    def test_civitai_find_online(self):
        resource = civitai_find_online(7240, version='Meina V11')
        assert resource.model_id == 7240
        assert resource.model_name == 'MeinaMix'
        assert resource.version_id == 119057
        assert resource.version_name == 'Meina V11'

        resource = civitai_find_online('明日方舟-安洁莉娜,Arknights-Angeline')
        assert resource.model_id == 5632
        assert resource.model_name == '明日方舟-安洁莉娜,Arknights-Angeline'
        assert resource.version_id == 6555
        assert resource.version_name == 'v1.0'
        assert resource.filename == '安洁莉娜5(头巾headband,三套衣服coat,multiple strap,swimsuit),.safetensors'
        assert resource.sha256 == '89A4D7CF20C76AFEE5AA4C4F28908A05D5E6BB6EB7726E5D319DE74731D09701'
        assert resource.is_primary
        assert resource.size == 37863532

        resource = civitai_find_online('toki/飛鳥馬トキ/时 (Blue Archive)', 'v1.4')
        assert resource.model_id == 121604
        assert resource.model_name == 'toki/飛鳥馬トキ/时 (Blue Archive)'
        assert resource.version_id == 153813
        assert resource.version_name == 'v1.4'
        assert resource.filename == 'toki_bluearchive.safetensors'
        assert resource.sha256 == '937404D0C14B2B14D87853DCCEDEC3EBF6ED0E79129E72659C2673873F6F0685'
        assert resource.is_primary
        assert resource.size == 14722160

        resource = civitai_find_online('toki/飛鳥馬トキ/时 (Blue Archive)', 'v1.4', '*.safetensors')
        assert resource.model_id == 121604
        assert resource.model_name == 'toki/飛鳥馬トキ/时 (Blue Archive)'
        assert resource.version_id == 153813
        assert resource.version_name == 'v1.4'
        assert resource.filename == 'toki_bluearchive.safetensors'
        assert resource.sha256 == '937404D0C14B2B14D87853DCCEDEC3EBF6ED0E79129E72659C2673873F6F0685'
        assert resource.is_primary
        assert resource.size == 14722160

        resource = civitai_find_online('toki/飛鳥馬トキ/时 (Blue Archive)', 'v1.4', '*.pt')
        assert resource.model_id == 121604
        assert resource.model_name == 'toki/飛鳥馬トキ/时 (Blue Archive)'
        assert resource.version_id == 153813
        assert resource.version_name == 'v1.4'
        assert resource.filename == 'toki_bluearchive.pt'
        assert resource.sha256 == '4CB82DDAE9CE0CBA475E89A083252F9179065B00BA233D0D9D8A44E323DDB53A'
        assert not resource.is_primary
        assert resource.size == 13141

        with pytest.raises(ResourceDuplicated):
            _ = civitai_find_online('toki/飛鳥馬トキ/时 (Blue Archive)', 'v1.4', '*')
        with pytest.raises(ModelFoundDuplicated):
            _ = civitai_find_online('Paimon (Genshin Impact)')

        resource = civitai_find_online('Paimon (Genshin Impact)', 'v1.4', creator='narugo1992')
        assert resource.model_id == 125187
        assert resource.model_name == 'Paimon (Genshin Impact)'
        assert resource.version_id == 156954
        assert resource.version_name == 'v1.4'
        assert resource.filename == 'paimon_genshin.safetensors'
        assert resource.sha256 == '3529E351565893CDB83DC8BDE9FF7F2E19D494B68A0AE6E43170F3C245016C8F'
        assert resource.is_primary
        assert resource.size == 14722160

    def test_civitai_search_online(self):
        results = civitai_search_online('kal_tsit_arknights', creator='narugo1992')
        assert len(results) == 1

        model = results[0]
        assert model.model_id == 121406
        assert model.model_name == "kal'tsit/ケルシー/凯尔希 (Arknights)"
        assert model.creator == 'narugo1992'
        assert repr(model) == '<Model name: "kal\'tsit/ケルシー/凯尔希 (Arknights)", id: 121406, creator: \'narugo1992\'>'

    def test_civitai_download(self):
        file = civitai_download('amiya arknights (old)')
        assert os.path.getsize(file) == 25515
        assert calculate_sha256(file) == '311279B35743D703DED68DF6ECC9F1E850A9F6976494CC8AF434DF3AA0E238A0'

        file = civitai_download('明日方舟-安洁莉娜,Arknights-Angeline')
        assert os.path.getsize(file) == 37863532
        assert calculate_sha256(file) == '89A4D7CF20C76AFEE5AA4C4F28908A05D5E6BB6EB7726E5D319DE74731D09701'

    def test_civitai_find_online_by_hash(self):
        resource = civitai_find_online('FB64F545')
        assert resource.model_id == 121986
        assert resource.model_name == 'mutsuki/浅黄ムツキ/睦月 (Blue Archive)'
        assert resource.version_id == 155681
        assert resource.version_name == 'v1.4'
        assert resource.creator == 'narugo1992'
        assert resource.filename == 'mutsuki_bluearchive.pt'
        assert resource.crc32 == 'FB64F545'
        assert not resource.is_primary
        assert resource.size == 13278

        resource = civitai_find_online('B42B09FF12')
        assert resource.model_id == 6755
        assert resource.model_name == 'Cetus-Mix'
        assert resource.version_id == 78676
        assert resource.version_name == 'V4'
        assert resource.creator == 'Eagelaxis'
        assert resource.filename == 'cetusMix_v4.safetensors'
        assert resource.crc32 == '838408E0'
        assert resource.is_primary
        assert resource.size == 3894258133



import os
from contextlib import contextmanager

import requests
from tqdm.auto import tqdm


class _FakeClass:
    def update(self, *args, **kwargs):
        pass


@contextmanager
def _with_tqdm(expected_size, desc, silent: bool = False):
    """
    Context manager that provides a tqdm progress bar for tracking the download progress.

    :param expected_size: The expected size of the file being downloaded.
    :type expected_size: int
    :param desc: The description of the progress bar.
    :type desc: str
    :param silent: Whether to silence the progress bar. If True, a fake progress bar is used. (default: False)
    :type silent: bool
    """
    if not silent:
        with tqdm(total=expected_size, unit='B', unit_scale=True, unit_divisor=1024, desc=desc) as pbar:
            yield pbar
    else:
        yield _FakeClass()


def download_file(url, filename, expected_size: int = None, desc=None, session=None, silent: bool = False, **kwargs):
    """
    Downloads a file from the given URL and saves it to the specified filename.

    :param url: The URL of the file to download.
    :type url: str
    :param filename: The filename to save the downloaded file to.
    :type filename: str
    :param expected_size: The expected size of the file in bytes. (default: None)
    :type expected_size: int
    :param desc: The description of the download progress. If not provided, the filename is used. (default: None)
    :type desc: str
    :param session: An existing requests Session object to use for the download. If not provided, a new Session object is created. (default: None)
    :type session: requests.Session
    :param silent: Whether to silence the progress bar. If True, no progress bar is displayed. (default: False)
    :type silent: bool
    :param kwargs: Additional keyword arguments to pass to the `srequest` function.
    :type kwargs: dict
    :returns: The filename of the downloaded file.
    :rtype: str
    """
    session = session or requests.session()
    response = session.get(url, stream=True, allow_redirects=True, **kwargs)
    expected_size = expected_size or response.headers.get('Content-Length', None)
    expected_size = int(expected_size) if expected_size is not None else expected_size

    desc = desc or os.path.basename(filename)
    directory = os.path.dirname(filename)
    if directory:
        os.makedirs(directory, exist_ok=True)

    with open(filename, 'wb') as f:
        with _with_tqdm(expected_size, desc, silent) as pbar:
            for chunk in response.iter_content(chunk_size=1024):
                f.write(chunk)
                pbar.update(len(chunk))

    actual_size = os.path.getsize(filename)
    if expected_size is not None and actual_size != expected_size:
        os.remove(filename)
        raise requests.exceptions.HTTPError(f"Downloaded file is not of expected size, "
                                            f"{expected_size} expected but {actual_size} found.")

    return filename



"""
Overview:
    Meta information for pycivitai package.
"""

#: Title of this project (should be `pycivitai`).
__TITLE__ = 'pycivitai'

#: Version of this project.
__VERSION__ = '0.2.0'

#: Short description of the project, will be included in ``setup.py``.
__DESCRIPTION__ = 'Python Client and Model Management for CivitAI.'

#: Author of this project.
__AUTHOR__ = 'narugo1992'

#: Email of the authors'.
__AUTHOR_EMAIL__ = 'narugo@126.com'
